# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0e
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import collections
import json
import tempfile

from awscli.customizations.cloudformation import exceptions
from awscli.customizations.cloudformation.deploy import DeployCommand
from awscli.customizations.cloudformation.deployer import Deployer
from awscli.customizations.cloudformation.yamlhelper import yaml_parse
from awscli.customizations.exceptions import ParamValidationError
from awscli.testutils import mock
from tests.unit.customizations.cloudformation import BaseYAMLTest


class FakeArgs:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __contains__(self, key):
        return key in self.__dict__


def get_example_template():
    return {"Parameters": {"Key1": "Value1"}, "Resources": {"Resource1": {}}}


ChangeSetResult = collections.namedtuple(
    "ChangeSetResult", ["changeset_id", "changeset_type"]
)


class TestDeployCommand(BaseYAMLTest):
    def setUp(self):
        super(TestDeployCommand, self).setUp()
        self.session = mock.Mock()
        self.session.get_scoped_config.return_value = {}
        self.parsed_args = FakeArgs(
            template_file='./foo',
            stack_name="some_stack_name",
            parameter_overrides=["Key1=Value1", "Key2=Value2"],
            no_execute_changeset=False,
            execute_changeset=True,
            disable_rollback=True,
            capabilities=None,
            role_arn=None,
            notification_arns=[],
            fail_on_empty_changeset=True,
            s3_bucket=None,
            s3_prefix="some prefix",
            kms_key_id="some kms key id",
            force_upload=True,
            tags=["tagkey1=tagvalue1"],
        )
        self.parsed_globals = FakeArgs(
            region="us-east-1", endpoint_url=None, verify_ssl=None
        )
        self.deploy_command = DeployCommand(self.session)

        self.deployer = Deployer(mock.Mock())
        self.deployer.create_and_wait_for_changeset = mock.Mock()
        self.deployer.execute_changeset = mock.Mock()
        self.deployer.wait_for_execute = mock.Mock()

    @mock.patch("awscli.customizations.cloudformation.deploy.yaml_parse")
    def test_command_invoked(self, mock_yaml_parse):
        """
        Tests that deploy method is invoked when command is run
        """
        fake_parameter_overrides = []
        fake_tags_dict = {"tagkey1": "tagvalue1"}
        fake_tags = [{"Key": "tagkey1", "Value": "tagvalue1"}]
        fake_parameters = "some return value"
        template_str = "some template"

        with tempfile.NamedTemporaryFile() as handle:
            file_path = handle.name

            open_mock = mock.mock_open()
            # Patch the file open method to return template string
            with mock.patch(
                "awscli.customizations.cloudformation.deploy.compat_open",
                open_mock(read_data=template_str),
            ) as open_mock:
                fake_template = get_example_template()
                mock_yaml_parse.return_value = fake_template

                self.deploy_command.deploy = mock.MagicMock()
                self.deploy_command.deploy.return_value = 0
                self.deploy_command.parse_key_value_arg = mock.Mock()
                self.deploy_command.parse_key_value_arg.side_effect = [
                    fake_parameter_overrides,
                    fake_tags_dict,
                ]
                self.deploy_command.merge_parameters = mock.MagicMock(
                    return_value=fake_parameters
                )

                self.parsed_args.template_file = file_path
                result = self.deploy_command._run_main(
                    self.parsed_args, parsed_globals=self.parsed_globals
                )
                self.assertEqual(0, result)
                open_mock.assert_called_once_with(file_path, "r")

                self.deploy_command.deploy.assert_called_once_with(
                    mock.ANY,
                    'some_stack_name',
                    mock.ANY,
                    fake_parameters,
                    None,
                    not self.parsed_args.no_execute_changeset,
                    None,
                    [],
                    None,
                    fake_tags,
                    True,
                    True,
                )

                self.deploy_command.parse_key_value_arg.assert_has_calls(
                    [
                        mock.call(
                            self.parsed_args.parameter_overrides,
                            "parameter-overrides",
                        ),
                        mock.call(self.parsed_args.tags, "tags"),
                    ]
                )

                self.deploy_command.merge_parameters.assert_called_once_with(
                    fake_template, fake_parameter_overrides
                )

                self.assertEqual(1, mock_yaml_parse.call_count)

    def test_invalid_template_file(self):
        self.parsed_args.template_file = "sometemplate"
        with self.assertRaises(exceptions.InvalidTemplatePathError):
            result = self.deploy_command._run_main(
                self.parsed_args, parsed_globals=self.parsed_globals
            )

    @mock.patch(
        'awscli.customizations.cloudformation.deploy.os.path.expanduser'
    )
    def test_expands_user_for_template_file(self, expanduser):
        contents = 'Resource:\n' '  Test:\n' '    Type: AWS::Foo::Bar\n'
        with tempfile.NamedTemporaryFile('w') as f:
            f.write(contents)
            f.flush()
            fake_expanduser_path = f.name + '.expanduser-file'
            with open(fake_expanduser_path, 'w') as f:
                f.write(contents)
        # The tempfile is now gone so if we weren't using os.path.expanduser()
        # in the code, this test would fail to load the file with an
        # InvalidTemplatePathError.
        expanduser.return_value = fake_expanduser_path
        parsed, template_str, _ = self.deploy_command.load_template_file(
            f.name
        )
        self.assertEqual(
            parsed, {'Resource': {'Test': {'Type': 'AWS::Foo::Bar'}}}
        )
        self.assertEqual(template_str, contents)

    @mock.patch('awscli.customizations.cloudformation.deploy.os.path.isfile')
    @mock.patch('awscli.customizations.cloudformation.deploy.yaml_parse')
    @mock.patch('awscli.customizations.cloudformation.deploy.os.path.getsize')
    def test_s3_upload_required_but_missing_bucket(
        self, mock_getsize, mock_yaml_parse, mock_isfile
    ):
        """
        Tests that large templates are detected prior to deployment
        """
        template_str = get_example_template()

        mock_getsize.return_value = 51201
        mock_isfile.return_value = True
        mock_yaml_parse.return_value = template_str
        open_mock = mock.mock_open()

        with mock.patch(
            "awscli.customizations.cloudformation.deploy.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            with self.assertRaises(exceptions.DeployBucketRequiredError):
                result = self.deploy_command._run_main(
                    self.parsed_args, parsed_globals=self.parsed_globals
                )

    @mock.patch('awscli.customizations.cloudformation.deploy.os.path.isfile')
    @mock.patch('awscli.customizations.cloudformation.deploy.yaml_parse')
    @mock.patch('awscli.customizations.cloudformation.deploy.os.path.getsize')
    @mock.patch(
        'awscli.customizations.cloudformation.deploy.DeployCommand.deploy'
    )
    @mock.patch('awscli.customizations.cloudformation.deploy.S3Uploader')
    def test_s3_uploader_is_configured_properly(
        self,
        s3UploaderMock,
        deploy_method_mock,
        mock_getsize,
        mock_yaml_parse,
        mock_isfile,
    ):
        """
        Tests that large templates are detected prior to deployment
        """
        bucket_name = "mybucket"
        template_str = get_example_template()

        mock_getsize.return_value = 1024
        mock_isfile.return_value = True
        mock_yaml_parse.return_value = template_str
        open_mock = mock.mock_open()

        with mock.patch(
            "awscli.customizations.cloudformation.deploy.compat_open",
            open_mock(read_data=template_str),
        ) as open_mock:
            self.parsed_args.s3_bucket = bucket_name
            s3UploaderObject = mock.Mock()
            s3UploaderMock.return_value = s3UploaderObject

            result = self.deploy_command._run_main(
                self.parsed_args, parsed_globals=self.parsed_globals
            )

            self.deploy_command.deploy.assert_called_once_with(
                mock.ANY,
                self.parsed_args.stack_name,
                mock.ANY,
                mock.ANY,
                None,
                not self.parsed_args.no_execute_changeset,
                None,
                [],
                s3UploaderObject,
                [{"Key": "tagkey1", "Value": "tagvalue1"}],
                True,
                True,
            )

            s3UploaderMock.assert_called_once_with(
                mock.ANY,
                bucket_name,
                self.parsed_args.s3_prefix,
                self.parsed_args.kms_key_id,
                self.parsed_args.force_upload,
            )

    def test_deploy_success(self):
        """
        Tests that we call the deploy command
        """

        stack_name = "stack_name"
        changeset_id = "some changeset"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = True
        changeset_type = "CREATE"
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        # Set the mock to return this fake changeset_id
        self.deployer.create_and_wait_for_changeset.return_value = (
            ChangeSetResult(changeset_id, changeset_type)
        )

        rc = self.deploy_command.deploy(
            self.deployer,
            stack_name,
            template,
            parameters,
            capabilities,
            execute_changeset,
            role_arn,
            notification_arns,
            s3_uploader,
            tags,
        )
        self.assertEqual(rc, 0)

        self.deployer.create_and_wait_for_changeset.assert_called_once_with(
            stack_name=stack_name,
            cfn_template=template,
            parameter_values=parameters,
            capabilities=capabilities,
            role_arn=role_arn,
            notification_arns=notification_arns,
            s3_uploader=s3_uploader,
            tags=tags,
        )

        # since execute_changeset is set to True, deploy() will execute changeset
        self.deployer.execute_changeset.assert_called_once_with(
            changeset_id, stack_name, False
        )
        self.deployer.wait_for_execute.assert_called_once_with(
            stack_name, changeset_type
        )

    def test_deploy_no_execute(self):
        stack_name = "stack_name"
        changeset_id = "some changeset"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = False
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        self.deployer.create_and_wait_for_changeset.return_value = (
            ChangeSetResult(changeset_id, "CREATE")
        )
        rc = self.deploy_command.deploy(
            self.deployer,
            stack_name,
            template,
            parameters,
            capabilities,
            execute_changeset,
            role_arn,
            notification_arns,
            s3_uploader,
            tags,
        )
        self.assertEqual(rc, 0)

        self.deployer.create_and_wait_for_changeset.assert_called_once_with(
            stack_name=stack_name,
            cfn_template=template,
            parameter_values=parameters,
            capabilities=capabilities,
            role_arn=role_arn,
            notification_arns=notification_arns,
            s3_uploader=s3_uploader,
            tags=tags,
        )

        # since execute_changeset is set to True, deploy() will execute changeset
        self.deployer.execute_changeset.assert_not_called()
        self.deployer.wait_for_execute.assert_not_called()

    def test_deploy_raise_exception(self):
        stack_name = "stack_name"
        changeset_id = "some changeset"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = True
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        s3_uploader = None
        tags = [{"Key": "key1", "Value": "val1"}]

        self.deployer.wait_for_execute.side_effect = RuntimeError("Some error")
        with self.assertRaises(RuntimeError):
            self.deploy_command.deploy(
                self.deployer,
                stack_name,
                template,
                parameters,
                capabilities,
                execute_changeset,
                role_arn,
                notification_arns,
                s3_uploader,
                tags,
            )

    def test_deploy_raises_exception_on_empty_changeset(self):
        stack_name = "stack_name"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = True
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]
        tags = []

        empty_changeset = exceptions.ChangeEmptyError(stack_name=stack_name)
        changeset_func = self.deployer.create_and_wait_for_changeset
        changeset_func.side_effect = empty_changeset
        with self.assertRaises(exceptions.ChangeEmptyError):
            self.deploy_command.deploy(
                self.deployer,
                stack_name,
                template,
                parameters,
                capabilities,
                execute_changeset,
                role_arn,
                notification_arns,
                None,
                tags,
                fail_on_empty_changeset=True,
            )

    def test_deploy_does_not_raise_exception_on_empty_changeset(self):
        stack_name = "stack_name"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = True
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]

        empty_changeset = exceptions.ChangeEmptyError(stack_name=stack_name)
        changeset_func = self.deployer.create_and_wait_for_changeset
        changeset_func.side_effect = empty_changeset
        self.deploy_command.deploy(
            self.deployer,
            stack_name,
            template,
            parameters,
            capabilities,
            execute_changeset,
            role_arn,
            notification_arns,
            s3_uploader=None,
            tags=[],
            fail_on_empty_changeset=False,
        )

    def test_deploy_empty_changeset_does_not_raise_exception_by_default(self):
        stack_name = "stack_name"
        parameters = ["a", "b"]
        template = "cloudformation template"
        capabilities = ["foo", "bar"]
        execute_changeset = True
        role_arn = "arn:aws:iam::1234567890:role"
        notification_arns = ["arn:aws:sns:region:1234567890:notify"]

        empty_changeset = exceptions.ChangeEmptyError(stack_name=stack_name)
        changeset_func = self.deployer.create_and_wait_for_changeset
        changeset_func.side_effect = empty_changeset
        self.deploy_command.deploy(
            self.deployer,
            stack_name,
            template,
            parameters,
            capabilities,
            execute_changeset,
            role_arn,
            notification_arns,
            s3_uploader=None,
            tags=[],
        )

    def test_parse_key_value_arg_success(self):
        """
        Tests that we can parse parameter arguments provided in proper format
        Expected format: ["Key=Value", "Key=Value"]
        :return:
        """
        argname = "parameter-overrides"
        data = ["Key1=Value1", 'Key2=[1,2,3]', 'Key3={"a":"val", "b": 2}']
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }

        result = self.deploy_command.parse_key_value_arg(data, argname)
        self.assertEqual(result, output)

        # Empty input should return empty output
        result = self.deploy_command.parse_key_value_arg([], argname)
        self.assertEqual(result, {})

    def test_parse_parameter_override_with_cf_data_format(self):
        """
        Tests that we can parse parameter arguments from file in
        CloudFormation parameters file format
        :return:
        """
        data = json.dumps(
            [
                {'ParameterKey': 'Key1', 'ParameterValue': 'Value1'},
                {'ParameterKey': 'Key2', 'ParameterValue': '[1,2,3]'},
                {
                    'ParameterKey': 'Key3',
                    'ParameterValue': '{"a":"val", "b": 2}',
                },
            ]
        )
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }
        result = self.deploy_command.parse_parameter_overrides(data)
        self.assertEqual(result, output)

    def test_validate_parameter_override_with_cf_data_format(self):
        """
        Tests that we through exception if have redundant keys in json
        CloudFormation parameters file format
        :return:
        """
        data = json.dumps(
            [
                {
                    'ParameterKey': 'Key1',
                    'ParameterValue': 'Value1',
                    'RedundantKey': 'foo',
                }
            ]
        )
        with self.assertRaises(ParamValidationError):
            self.deploy_command.parse_parameter_overrides(data)

    def test_parse_parameter_override_with_codepipeline_data_format(self):
        """
        Tests that we can parse parameter arguments from file in
        CodePipeline parameters file format
        :return:
        """
        data = json.dumps(
            {
                'Parameters': {
                    "Key1": "Value1",
                    "Key2": '[1,2,3]',
                    "Key3": '{"a":"val", "b": 2}',
                }
            }
        )
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }
        result = self.deploy_command.parse_parameter_overrides(data)
        self.assertEqual(result, output)

    def test_parse_parameter_override_with_deploy_data_format_from_file(self):
        """
        Tests that we can parse parameter arguments from file in
        deploy command parameters file format
        :return:
        """
        data = json.dumps(
            ['Key1=Value1', 'Key2=[1,2,3]', 'Key3={"a":"val", "b": 2}']
        )
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }
        result = self.deploy_command.parse_parameter_overrides(data)
        self.assertEqual(result, output)

    def test_parse_parameter_override_with_inline_json(self):
        data = [
            json.dumps(
                ['Key1=Value1', 'Key2=[1,2,3]', 'Key3={"a":"val", "b": 2}']
            )
        ]
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }
        result = self.deploy_command.parse_parameter_overrides(data)
        self.assertEqual(result, output)

    def test_parse_parameter_override_with_deploy_data_format(self):
        """
        Tests that we can parse parameter arguments in
        deploy command parameters command line format
        :return:
        """
        data = ['Key1=Value1', 'Key2=[1,2,3]', 'Key3={"a":"val", "b": 2}']
        output = {
            "Key1": "Value1",
            "Key2": '[1,2,3]',
            "Key3": '{"a":"val", "b": 2}',
        }
        result = self.deploy_command.parse_parameter_overrides(data)
        self.assertEqual(result, output)

    def test_parse_key_value_arg_invalid_input(self):
        # non-list input
        argname = "parameter-overrides"
        with self.assertRaises(exceptions.InvalidKeyValuePairArgumentError):
            self.deploy_command.parse_key_value_arg("hello=world", argname)

        # missing equal to sign
        with self.assertRaises(exceptions.InvalidKeyValuePairArgumentError):
            self.deploy_command.parse_key_value_arg(["hello world"], argname)

    def test_merge_parameters_success(self):
        """
        Tests that we can merge parameters specified in CloudFormation template
        with override values specified as commandline arguments
        """
        template = {
            "Parameters": {
                "Key1": {"Type": "String"},
                "Key2": {"Type": "String"},
                "Key3": "Something",
                "Key4": {"Type": "Number"},
                "KeyWithDefaultValue": {
                    "Type": "String",
                    "Default": "something",
                },
                "KeyWithDefaultValueButOverridden": {
                    "Type": "String",
                    "Default": "something",
                },
            }
        }
        overrides = {
            "Key1": "Value1",
            "Key3": "Value3",
            "KeyWithDefaultValueButOverridden": "Value4",
        }

        expected_result = [
            # Overriden values
            {"ParameterKey": "Key1", "ParameterValue": "Value1"},
            {"ParameterKey": "Key3", "ParameterValue": "Value3"},
            # Parameter contains default value, but overridden with new value
            {
                "ParameterKey": "KeyWithDefaultValueButOverridden",
                "ParameterValue": "Value4",
            },
            # non-overriden values
            {"ParameterKey": "Key2", "UsePreviousValue": True},
            {"ParameterKey": "Key4", "UsePreviousValue": True},
            # Parameter with default value but NOT overridden.
            # Use previous value, but this gets removed later when we are creating stack for first time
            {"ParameterKey": "KeyWithDefaultValue", "UsePreviousValue": True},
        ]

        self.assertItemsEqual(
            self.deploy_command.merge_parameters(template, overrides),
            expected_result,
        )

    def test_merge_parameters_success_nothing_to_override(self):
        """
        Tests that we can merge parameters specified in CloudFormation template
        with override values specified as commandline arguments
        """
        template = {
            "Parameters": {
                "Key1": {"Type": "String"},
                "Key2": {"Type": "String"},
                "Key3": "Something",
                "Key4": {"Type": "Number"},
            }
        }
        overrides = {
            # Key5 is not in the template. We will simply skip this key
            "Key5": "Value5"
        }

        expected_result = [
            {"ParameterKey": "Key1", "UsePreviousValue": True},
            {"ParameterKey": "Key2", "UsePreviousValue": True},
            {"ParameterKey": "Key3", "UsePreviousValue": True},
            {"ParameterKey": "Key4", "UsePreviousValue": True},
        ]

        self.assertItemsEqual(
            self.deploy_command.merge_parameters(template, overrides),
            expected_result,
        )

        # Parameters definition is empty. Nothing to override
        result = self.deploy_command.merge_parameters(
            {"Parameters": {}}, overrides
        )
        self.assertEqual(result, [])

    def test_merge_parameters_invalid_input(self):
        # Template does not contain "Parameters" definition
        result = self.deploy_command.merge_parameters({}, {"Key": "Value"})
        self.assertEqual(result, [])

        # Parameters definition is invalid
        result = self.deploy_command.merge_parameters(
            {"Parameters": "foo"}, {"Key": "Value"}
        )
        self.assertEqual(result, [])



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0e
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.customizations.s3.s3 import add_s3, awscli_initialize
from awscli.testutils import BaseAWSCommandParamsTest, mock, unittest


class AWSInitializeTest(unittest.TestCase):
    """
    This test ensures that all events are correctly registered such that
    all of the commands can be run.
    """

    def setUp(self):
        self.cli = mock.Mock()

    def test_initialize(self):
        awscli_initialize(self.cli)
        reference = []
        reference.append("building-command-table.main")
        reference.append("building-command-table.s3_sync")
        for arg in self.cli.register.call_args_list:
            self.assertIn(arg[0][0], reference)


class CreateTablesTest(unittest.TestCase):
    def test_s3(self):
        """
        Ensures that the table for the service was created properly.
        Also ensures the original s3 service is renamed to ``s3api``.
        """
        s3_service = mock.Mock()
        s3_service.name = 's3'
        self.services = {'s3': s3_service}
        add_s3(self.services, True)
        orig_service = self.services.pop('s3api')
        self.assertEqual(orig_service, s3_service)
        for service in self.services.keys():
            self.assertIn(service, ['s3'])


class TestS3(BaseAWSCommandParamsTest):
    def test_too_few_args(self):
        stderr = self.run_cmd('s3', expected_rc=252)[1]
        self.assertIn(
            ("usage: aws [options] s3 " "<subcommand> [parameters]"), stderr
        )
        self.assertIn('too few arguments', stderr)


if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0e
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import sys
import tempfile
from io import StringIO

from awscli.customizations.cloudformation.artifact_exporter import Template
from awscli.customizations.cloudformation.package import PackageCommand
from awscli.customizations.cloudformation.yamlhelper import yaml_dump
from awscli.testutils import mock
from tests.unit.customizations.cloudformation import BaseYAMLTest


class FakeArgs:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __contains__(self, key):
        return key in self.__dict__


def get_example_template():
    return {"Parameters": {"Key1": "Value1"}, "Resources": {"Resource1": {}}}


class TestPackageCommand(BaseYAMLTest):
    def setUp(self):
        super(TestPackageCommand, self).setUp()
        self.session = mock.Mock()
        self.session.get_scoped_config.return_value = {}
        self.parsed_args = FakeArgs(
            template_file='./foo',
            s3_bucket="s3bucket",
            s3_prefix="s3prefix",
            kms_key_id="kmskeyid",
            output_template_file="./oputput",
            use_json=False,
            force_upload=False,
            metadata=None,
        )
        self.parsed_globals = FakeArgs(
            region="us-east-1", endpoint_url=None, verify_ssl=None
        )
        self.package_command = PackageCommand(self.session)

    @mock.patch("awscli.customizations.cloudformation.package.yaml_dump")
    def test_main(self, mock_yaml_dump):
        exported_template_str = "hello"

        self.package_command.write_output = mock.Mock()
        self.package_command._export = mock.Mock()
        mock_yaml_dump.return_value = exported_template_str

        # Create a temporary file and make this my template
        with tempfile.NamedTemporaryFile() as handle:
            for use_json in (False, True):
                filename = handle.name
                self.parsed_args.template_file = filename
                self.parsed_args.use_json = use_json

                rc = self.package_command._run_main(
                    self.parsed_args, self.parsed_globals
                )
                self.assertEqual(rc, 0)

                self.package_command._export.assert_called_once_with(
                    filename, use_json
                )
                self.package_command.write_output.assert_called_once_with(
                    self.parsed_args.output_template_file, mock.ANY
                )

                self.package_command._export.reset_mock()
                self.package_command.write_output.reset_mock()

    def test_main_error(self):
        self.package_command._export = mock.Mock()
        self.package_command._export.side_effect = RuntimeError()

        # Create a temporary file and make this my template
        with tempfile.NamedTemporaryFile() as handle:
            filename = handle.name
            self.parsed_args.template_file = filename

            with self.assertRaises(RuntimeError):
                self.package_command._run_main(
                    self.parsed_args, self.parsed_globals
                )

    @mock.patch("awscli.customizations.cloudformation.package.sys.stdout")
    def test_write_output_to_stdout(self, stdoutmock):
        data = "some data"
        filename = None

        self.package_command.write_output(filename, data)
        stdoutmock.write.assert_called_once_with(data)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from awscli.autocomplete import parser
from awscli.autoprompt.output import OutputGetter
from awscli.clidriver import create_clidriver
from awscli.testutils import unittest
from tests.unit.autocomplete import InMemoryIndex


class TestOutputGetter(unittest.TestCase):
    def setUp(self):
        index = InMemoryIndex(
            {
                'command_names': {
                    '': [('aws', None)],
                    'aws': [
                        ('s3api', 'Amazon Simple Storage Service'),
                        ('ec2', None),
                    ],
                    'aws.s3api': [('get-object', None)],
                    'aws.ec2': [('bundle-instance', None)],
                },
                'arg_names': {
                    '': {
                        'aws': ['query', 'output'],
                    },
                    'aws.s3api': {
                        'get-object': [],
                    },
                    'aws.ec2': {
                        'bundle-instance': [],
                    },
                },
                'arg_data': {
                    '': {
                        'aws': {
                            'query': (
                                'query',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                            'output': (
                                'output',
                                'string',
                                'aws',
                                '',
                                None,
                                False,
                                False,
                            ),
                        }
                    },
                    'aws.s3api': {'get-object': {}},
                    'aws.ec2': {
                        'bundle-instance': {},
                    },
                },
            }
        )
        self.parser = parser.CLIParser(index)
        self.driver = create_clidriver()
        self.driver.session.set_config_variable('output', 'json')
        self.base_output_getter = OutputGetter(self.driver)

    def test_get_output_content(self):
        parsed = self.parser.parse('aws s3api get-object')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('"Body":', content)

    def test_can_return_no_output(self):
        parsed = self.parser.parse('aws s3')
        content = self.base_output_getter.get_output(parsed)
        self.assertEqual('No output', content)

    def test_can_return_json(self):
        parsed = self.parser.parse('aws s3api get-object --output json ')
        content = json.loads(self.base_output_getter.get_output(parsed))
        self.assertIn('Body', content)

    def test_can_take_output_format_wo_trailing_space(self):
        parsed = self.parser.parse('aws s3api get-object --output table')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('------+------', content)

    def test_use_session_output_value_on_partial_input(self):
        parsed = self.parser.parse('aws s3api get-object --output yam')
        content = json.loads(self.base_output_getter.get_output(parsed))
        self.assertIn('Body', content)

    def test_can_use_session_output_format(self):
        parsed = self.parser.parse('aws s3api get-object')
        content = json.loads(self.base_output_getter.get_output(parsed))
        self.assertIn('Body', content)

    def test_can_return_error_message_on_invalid_output_format(self):
        parsed = self.parser.parse('aws s3api get-object --output poem ')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('Bad value for --output: poem', content)

    def test_can_return_error_message_on_invalid_query(self):
        parsed = self.parser.parse('aws s3api get-object --query A[}')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('Bad value for --query: A[}', content)

    def test_can_handle_open_brackets(self):
        parsed = self.parser.parse('aws s3api get-object --query Metadata.[')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('KeyName', content)

    def test_can_handle_trailing_dot(self):
        parsed = self.parser.parse('aws s3api get-object --query Metadata.')
        content = self.base_output_getter.get_output(parsed)
        self.assertIn('KeyName', content)

    def test_not_show_omap_in_yaml_output(self):
        parsed = self.parser.parse('aws s3api get-object --output yaml')
        content = self.base_output_getter.get_output(parsed)
        self.assertNotIn('omap', content)

    def test_yaml_output_can_parse_datetime(self):
        parsed = self.parser.parse(
            'aws ec2 bundle-instance --output yaml '
            '--query BundleTask.StartTime'
        )
        content = self.base_output_getter.get_output(parsed)
        self.assertEqual('"1970-01-01T00:00:00"\n', content)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from argparse import Namespace

from awscli.compat import StringIO
from awscli.customizations.configure.list import ConfigureListCommand
from awscli.testutils import mock, unittest

from . import FakeSession


class TestConfigureListCommand(unittest.TestCase):
    def test_configure_list_command_nothing_set(self):
        # Test the case where the user only wants to change a single_value.
        session = FakeSession(
            all_variables={'config_file': '/config/location'}
        )
        session.full_config = {
            'profiles': {'default': {'region': 'AWS_DEFAULT_REGION'}}
        }
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegex(rendered, r'profile\s+: <not set>')
        self.assertRegex(rendered, r'access_key\s+: <not set>')
        self.assertRegex(rendered, r'secret_key\s+: <not set>')
        self.assertRegex(rendered, r'region\s+: <not set>')

    def test_configure_from_env(self):
        env_vars = {'profile': 'myprofilename'}
        session = FakeSession(
            all_variables={'config_file': '/config/location'},
            environment_vars=env_vars,
        )
        session.session_var_map = {'profile': (None, "PROFILE_ENV_VAR")}
        session.full_config = {
            'profiles': {'default': {'region': 'AWS_DEFAULT_REGION'}}
        }
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegex(
            rendered, r'profile\s+: myprofilename\s+: env\s+: PROFILE_ENV_VAR'
        )

    def test_configure_from_config_file(self):
        config_file_vars = {'region': 'us-west-2'}
        session = FakeSession(
            all_variables={'config_file': '/config/location'},
            config_file_vars=config_file_vars,
        )
        session.session_var_map = {'region': ('region', "AWS_DEFAULT_REGION")}
        session.full_config = {
            'profiles': {'default': {'region': 'AWS_DEFAULT_REGION'}}
        }
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegex(
            rendered,
            r'region\s+: us-west-2\s+: config-file\s+: /config/location',
        )

    def test_configure_from_multiple_sources(self):
        # Here the profile is from an env var, the
        # region is from the config file, and the credentials
        # are from an iam-role.
        env_vars = {'profile': 'myprofilename'}
        config_file_vars = {'region': 'us-west-2'}
        credentials = mock.Mock()
        credentials.access_key = 'access_key'
        credentials.secret_key = 'secret_key'
        credentials.method = 'iam-role'
        session = FakeSession(
            all_variables={'config_file': '/config/location'},
            environment_vars=env_vars,
            config_file_vars=config_file_vars,
            credentials=credentials,
        )
        session.session_var_map = {
            'region': ('region', 'AWS_DEFAULT_REGION'),
            'profile': ('profile', 'AWS_DEFAULT_PROFILE'),
        }
        session.full_config = {
            'profiles': {'default': {'region': 'AWS_DEFAULT_REGION'}}
        }
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=None)
        rendered = stream.getvalue()
        # The profile came from an env var.
        self.assertRegex(
            rendered,
            r'profile\s+: myprofilename\s+: env\s+: AWS_DEFAULT_PROFILE',
        )
        # The region came from the config file.
        self.assertRegex(
            rendered,
            r'region\s+: us-west-2\s+: config-file\s+: /config/location',
        )
        # The credentials came from an IAM role.  Note how we're
        # also checking that the access_key/secret_key are masked
        # with '*' chars except for the last 4 chars.
        self.assertRegex(rendered, r'access_key\s+: \*+_key\s+: iam-role')
        self.assertRegex(rendered, r'secret_key\s+: \*+_key\s+: iam-role')

    def test_configure_region_from_imds(self):
        session = FakeSession(all_variables={'region': 'from-imds'})
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegex(rendered, r'region\s+: from-imds\s+: imds')

    def test_configure_from_args(self):
        parsed_globals = Namespace(profile='foo')
        env_vars = {'profile': 'myprofilename'}
        session = FakeSession(
            all_variables={'config_file': '/config/location'},
            profile='foo',
            environment_vars=env_vars,
        )
        session.session_var_map = {'profile': (None, ['AWS_PROFILE'])}
        session.full_config = {'profiles': {'foo': {'region': 'AWS_REGION'}}}
        stream = StringIO()
        self.configure_list = ConfigureListCommand(session, stream)
        self.configure_list(args=[], parsed_globals=parsed_globals)
        rendered = stream.getvalue()
        self.assertRegex(rendered, r'profile\s+: foo\s+: manual\s+: --profile')



#!/usr/bin/env python
# Copyright 2012-2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import BytesIO
from awscli.testutils import BaseAWSCommandParamsTest, FileCreator


class TestStreamingOutput(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestStreamingOutput, self).setUp()
        self.files = FileCreator()

    def tearDown(self):
        super(TestStreamingOutput, self).tearDown()
        self.files.remove_all()

    def test_get_media_streaming_output(self):
        cmdline = (
            'kinesis-video-media get-media --stream-name test-stream '
            '--start-selector StartSelectorType=EARLIEST %s'
        )
        self.parsed_response = {
            'ContentType': 'video/webm',
            'Payload': BytesIO(b'testbody'),
        }
        outpath = self.files.full_path('outfile')
        params = {
            'StartSelector': {'StartSelectorType': 'EARLIEST'},
            'StreamName': 'test-stream',
        }
        self.assert_params_for_cmd(cmdline % outpath, params)
        with open(outpath, 'rb') as outfile:
            self.assertEqual(outfile.read(), b'testbody')



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.credentials
import botocore.session
from botocore.exceptions import ProfileNotFound, UnknownCredentialError

from awscli.customizations.sso import inject_json_file_cache
from awscli.testutils import mock, unittest


class TestInjectJSONFileCache(unittest.TestCase):
    def setUp(self):
        self.mock_sso_provider = mock.Mock(
            spec=botocore.credentials.SSOProvider
        )
        self.mock_resolver = mock.Mock(botocore.credentials.CredentialResolver)
        self.mock_resolver.get_provider.return_value = self.mock_sso_provider
        self.session = mock.Mock(spec=botocore.session.Session)
        self.session.get_component.return_value = self.mock_resolver

    def test_inject_json_file_cache(self):
        inject_json_file_cache(self.session, event_name='session-initialized')
        self.session.get_component.assert_called_with('credential_provider')
        self.assertIsInstance(
            self.mock_sso_provider.cache,
            botocore.credentials.JSONFileCache,
        )

    def test_profile_not_found_is_not_propagated(self):
        self.session.get_component.side_effect = ProfileNotFound(
            profile='unknown'
        )
        try:
            inject_json_file_cache(
                self.session, event_name='session-initialized'
            )
        except ProfileNotFound:
            self.fail('ProfileNotFound should not have been raised.')

    def test_provider_not_found_error_is_not_propagated(self):
        self.mock_resolver.get_provider.side_effect = UnknownCredentialError(
            name='sso'
        )
        try:
            inject_json_file_cache(
                self.session, event_name='session-initialized'
            )
        except UnknownCredentialError:
            self.fail('UnknownCredentialError should not have been raised.')



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import shutil
import tempfile

from botocore.compat import OrderedDict

from awscli.customizations.eks.kubeconfig import (
    Kubeconfig,
    KubeconfigInaccessableError,
    KubeconfigLoader,
    KubeconfigValidator,
    KubeconfigWriter,
    _get_new_kubeconfig_content,
)
from awscli.testutils import mock, unittest
from tests.functional.eks.test_util import get_testdata


class TestKubeconfigWriter(unittest.TestCase):
    def setUp(self):
        self._writer = KubeconfigWriter()

    def test_write_order(self):
        content = OrderedDict(
            [("current-context", "context"), ("apiVersion", "v1")]
        )
        file_to_write = tempfile.NamedTemporaryFile(mode='w').name
        self.addCleanup(os.remove, file_to_write)

        config = Kubeconfig(file_to_write, content)
        self._writer.write_kubeconfig(config)

        with open(file_to_write) as stream:
            self.assertMultiLineEqual(
                stream.read(), "current-context: context\n" "apiVersion: v1\n"
            )

    def test_write_makedirs(self):
        content = OrderedDict(
            [("current-context", "context"), ("apiVersion", "v1")]
        )
        containing_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, containing_dir)
        config_path = os.path.join(containing_dir, "dir1", "dir2", "dir3")

        config = Kubeconfig(config_path, content)
        self._writer.write_kubeconfig(config)

        with open(config_path) as stream:
            self.assertMultiLineEqual(
                stream.read(), "current-context: context\n" "apiVersion: v1\n"
            )

    def test_write_directory(self):
        content = OrderedDict(
            [("current-context", "context"), ("apiVersion", "v1")]
        )
        containing_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, containing_dir)

        config = Kubeconfig(containing_dir, content)
        self.assertRaises(
            KubeconfigInaccessableError, self._writer.write_kubeconfig, config
        )


class TestKubeconfigLoader(unittest.TestCase):
    def setUp(self):
        # This mock validator allows all kubeconfigs
        self._validator = mock.Mock(spec=KubeconfigValidator)
        self._loader = KubeconfigLoader(self._validator)

        self._temp_directory = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self._temp_directory)

    def _clone_config(self, config):
        """
        Copies the testdata named config into the temp directory,
        Returns the new path

        :param config: The name of the testdata to copy
        :type config: str
        """
        old_path = os.path.abspath(get_testdata(config))
        new_path = os.path.join(self._temp_directory, config)
        shutil.copy2(old_path, new_path)
        return new_path

    def test_load_simple(self):
        simple_path = self._clone_config("valid_simple")
        content = OrderedDict(
            [
                ("apiVersion", "v1"),
                (
                    "clusters",
                    [
                        OrderedDict(
                            [
                                (
                                    "cluster",
                                    OrderedDict([("server", "simple")]),
                                ),
                                ("name", "simple"),
                            ]
                        )
                    ],
                ),
                ("contexts", None),
                ("current-context", "simple"),
                ("kind", "Config"),
                ("preferences", OrderedDict()),
                ("users", None),
            ]
        )
        loaded_config = self._loader.load_kubeconfig(simple_path)
        self.assertEqual(loaded_config.content, content)
        self._validator.validate_config.assert_called_with(
            Kubeconfig(simple_path, content)
        )

    def test_load_noexist(self):
        no_exist_path = os.path.join(
            self._temp_directory, "this_does_not_exist"
        )
        loaded_config = self._loader.load_kubeconfig(no_exist_path)
        self.assertEqual(loaded_config.content, _get_new_kubeconfig_content())
        self._validator.validate_config.assert_called_with(
            Kubeconfig(no_exist_path, _get_new_kubeconfig_content())
        )

    def test_load_empty(self):
        empty_path = self._clone_config("valid_empty_existing")
        loaded_config = self._loader.load_kubeconfig(empty_path)
        self.assertEqual(loaded_config.content, _get_new_kubeconfig_content())
        self._validator.validate_config.assert_called_with(
            Kubeconfig(empty_path, _get_new_kubeconfig_content())
        )

    def test_load_directory(self):
        current_directory = self._temp_directory
        self.assertRaises(
            KubeconfigInaccessableError,
            self._loader.load_kubeconfig,
            current_directory,
        )
        self._validator.validate_config.assert_not_called()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
Defines SSO-specific configure commands, to be used with the `aws configure`
command.

The main reason it lives in its own separate module instead of in
`awscli/customizations/configure/sso.py` is so that these commands can be
referenced without importing `awscli/customizations/configure/sso.py`,
which imports from `prompt_toolkit`. Importing from `prompt_toolkit` has
historically increased command execution times.

This separation helps us limit our imports from `prompt_toolkit` to when it
is actually needed, improving execution time across most commands.
"""

import logging
import os

import colorama
from botocore import UNSIGNED
from botocore.config import Config
from botocore.configprovider import ConstantProvider
from botocore.exceptions import ProfileNotFound

from awscli.customizations.configure import (
    get_section_header,
    profile_to_section,
)
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.sso.utils import (
    LOGIN_ARGS,
    BaseSSOCommand,
    PrintOnlyHandler,
    do_sso_login,
)
from awscli.customizations.utils import uni_print
from awscli.formatter import CLI_OUTPUT_FORMATS

logger = logging.getLogger(__name__)

_CMD_PROMPT_USAGE = (
    'To keep an existing value, hit enter when prompted for the value. When '
    'you are prompted for information, the current value will be displayed in '
    '[brackets]. If the config item has no value, it is displayed as '
    '[None] or omitted entirely.\n\n'
)
_CONFIG_EXTRA_INFO = (
    'Note: The configuration is saved in the shared configuration file. '
    'By default, ``~/.aws/config``. For more information, see the '
    '"Configuring the AWS CLI to use AWS IAM Identity Center" section in the '
    'AWS CLI User Guide:'
    '\n\nhttps://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sso.html'
)


def display_account(account):
    """Converts an SSO account response into a display string.

    All fields should be present in the API response but we've seen some
    cases where the account name or email address is not set. Considering
    we only need the account name and email address for display purposes
    we should be defensive just in case they don't come back.
    """
    if 'accountName' not in account and 'emailAddress' not in account:
        account_template = '{accountId}'
    elif 'emailAddress' not in account:
        account_template = '{accountName} ({accountId})'
    elif 'accountName' not in account:
        account_template = '{emailAddress} ({accountId})'
    else:
        account_template = '{accountName}, {emailAddress} ({accountId})'
    return account_template.format(**account)


def get_account_sorting_key(account):
    only_account_id = (
        'accountName' not in account and 'emailAddress' not in account
    )
    for key in ('accountName', 'emailAddress', 'accountId'):
        value = account.get(key, None)
        if value is not None:
            return (only_account_id, value.lower())
    return (only_account_id, None)


class BaseSSOConfigurationCommand(BaseSSOCommand):
    def __init__(self, session, prompter=None, config_writer=None):
        super().__init__(session)
        self._prompter = prompter
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer
        self._sso_sessions = self._session.full_config.get('sso_sessions', {})
        # Initialize self._sso_session_prompter to None. It will be
        # initialized lazily during command execution.
        self._sso_session_prompter = None

    def _init_prompt_toolkit(self):
        from awscli.customizations.configure.sso import (
            PTKPrompt,
            SSOSessionConfigurationPrompter,
        )

        if self._prompter is None:
            self._prompter = PTKPrompt()

        self._sso_session_prompter = SSOSessionConfigurationPrompter(
            botocore_session=self._session,
            prompter=self._prompter,
        )

    def _write_sso_configuration(self):
        self._update_section(
            section_header=get_section_header(
                'sso-session', self._sso_session_prompter.sso_session
            ),
            new_values=self._sso_session_prompter.sso_session_config,
        )

    def _update_section(self, section_header, new_values):
        config_path = self._session.get_config_variable('config_file')
        config_path = os.path.expanduser(config_path)
        new_values['__section__'] = section_header
        self._config_writer.update_config(new_values, config_path)

    def _run_main(self, parsed_args, parsed_globals):
        self._init_prompt_toolkit()


class ConfigureSSOCommand(BaseSSOConfigurationCommand):
    NAME = 'sso'
    SYNOPSIS = 'aws configure sso [--profile profile-name]'
    DESCRIPTION = (
        'The ``aws configure sso`` command interactively prompts for the '
        'configuration values required to create a profile that sources '
        'temporary AWS credentials from AWS IAM Identity Center.\n\n'
        f'{_CMD_PROMPT_USAGE}'
        'When providing the ``--profile`` parameter the named profile '
        'will be created or updated. When a profile is not explicitly set '
        'the profile name will be prompted for.\n\n'
        f'{_CONFIG_EXTRA_INFO}'
    )
    # TODO: Add CLI parameters to skip prompted values, --start-url, etc.
    ARG_TABLE = LOGIN_ARGS

    def __init__(
        self,
        session,
        prompter=None,
        selector=None,
        config_writer=None,
        sso_token_cache=None,
        sso_login=None,
    ):
        super().__init__(
            session, prompter=prompter, config_writer=config_writer
        )
        self._selector = selector
        if sso_login is None:
            sso_login = do_sso_login
        self._sso_login = sso_login
        self._sso_token_cache = sso_token_cache

        self._new_profile_config_values = {}
        self._original_profile_name = self._session.profile
        try:
            self._profile_config = self._session.get_scoped_config()
        except ProfileNotFound:
            self._profile_config = {}

    def _init_prompt_toolkit(self):
        super()._init_prompt_toolkit()
        if self._selector is None:
            from awscli.customizations.wizard.ui.selectmenu import select_menu

            self._selector = select_menu
        self._set_sso_session_if_configured_in_profile()

    def _set_sso_session_if_configured_in_profile(self):
        if 'sso_session' in self._profile_config:
            self._sso_session_prompter.sso_session = self._profile_config[
                'sso_session'
            ]

    def _handle_single_account(self, accounts):
        sso_account_id = accounts[0]['accountId']
        single_account_msg = 'The only AWS account available to you is: {}\n'
        uni_print(single_account_msg.format(sso_account_id))
        return sso_account_id

    def _handle_multiple_accounts(self, accounts):
        available_accounts_msg = (
            'There are {} AWS accounts available to you.\n'
        )
        uni_print(available_accounts_msg.format(len(accounts)))
        sorted_accounts = sorted(accounts, key=get_account_sorting_key)
        selected_account = self._selector(
            sorted_accounts, display_format=display_account
        )
        sso_account_id = selected_account['accountId']
        return sso_account_id

    def _get_all_accounts(self, sso, sso_token):
        paginator = sso.get_paginator('list_accounts')
        results = paginator.paginate(accessToken=sso_token['accessToken'])
        return results.build_full_result()

    def _prompt_for_account(self, sso, sso_token):
        accounts = self._get_all_accounts(sso, sso_token)['accountList']
        if not accounts:
            raise RuntimeError('No AWS accounts are available to you.')
        if len(accounts) == 1:
            sso_account_id = self._handle_single_account(accounts)
        else:
            sso_account_id = self._handle_multiple_accounts(accounts)
        uni_print(f'Using the account ID {sso_account_id}\n')
        self._new_profile_config_values['sso_account_id'] = sso_account_id
        return sso_account_id

    def _handle_single_role(self, roles):
        sso_role_name = roles[0]['roleName']
        available_roles_msg = 'The only role available to you is: {}\n'
        uni_print(available_roles_msg.format(sso_role_name))
        return sso_role_name

    def _handle_multiple_roles(self, roles):
        available_roles_msg = 'There are {} roles available to you.\n'
        uni_print(available_roles_msg.format(len(roles)))
        sorted_roles = sorted(roles, key=lambda x: x['roleName'].lower())
        role_names = [r['roleName'] for r in sorted_roles]
        sso_role_name = self._selector(role_names)
        return sso_role_name

    def _get_all_roles(self, sso, sso_token, sso_account_id):
        paginator = sso.get_paginator('list_account_roles')
        results = paginator.paginate(
            accountId=sso_account_id, accessToken=sso_token['accessToken']
        )
        return results.build_full_result()

    def _prompt_for_role(self, sso, sso_token, sso_account_id):
        roles = self._get_all_roles(sso, sso_token, sso_account_id)['roleList']
        if not roles:
            error_msg = 'No roles are available for the account {}'
            raise RuntimeError(error_msg.format(sso_account_id))
        if len(roles) == 1:
            sso_role_name = self._handle_single_role(roles)
        else:
            sso_role_name = self._handle_multiple_roles(roles)
        uni_print(f'Using the role name "{sso_role_name}"\n')
        self._new_profile_config_values['sso_role_name'] = sso_role_name
        return sso_role_name

    def _prompt_for_profile(self, sso_account_id=None, sso_role_name=None):
        from awscli.customizations.configure.sso import RequiredInputValidator

        if self._original_profile_name:
            profile_name = self._original_profile_name
        else:
            text = 'Profile name'
            default_profile = None
            if sso_account_id and sso_role_name:
                default_profile = f'{sso_role_name}-{sso_account_id}'
            validator = RequiredInputValidator(default_profile)
            profile_name = self._prompter.get_value(
                default_profile, text, validator=validator
            )
        return profile_name

    def _prompt_for_cli_default_region(self):
        # TODO: figure out a way to get a list of reasonable client regions
        return self._prompt_for_profile_config(
            'region', 'Default client Region'
        )

    def _prompt_for_cli_output_format(self):
        return self._prompt_for_profile_config(
            'output',
            'CLI default output format (json if not specified)',
            completions=list(CLI_OUTPUT_FORMATS.keys()),
        )

    def _prompt_for_profile_config(self, config_name, text, completions=None):
        current_value = self._profile_config.get(config_name)

        new_value = self._prompter.get_value(
            current_value,
            text,
            completions=completions,
        )
        if new_value:
            self._new_profile_config_values[config_name] = new_value
        return new_value

    def _unset_session_profile(self):
        config_store = self._session.get_component('config_store')
        config_store.set_config_provider('profile', ConstantProvider(None))

    def _run_main(self, parsed_args, parsed_globals):
        super()._run_main(parsed_args, parsed_globals)
        self._unset_session_profile()
        on_pending_authorization = None
        if parsed_args.no_browser:
            on_pending_authorization = PrintOnlyHandler()
        sso_registration_args = self._prompt_for_sso_registration_args()
        sso_token = self._sso_login(
            self._session,
            parsed_globals=parsed_globals,
            token_cache=self._sso_token_cache,
            on_pending_authorization=on_pending_authorization,
            use_device_code=parsed_args.use_device_code,
            **sso_registration_args,
        )

        client_config = Config(
            signature_version=UNSIGNED,
            region_name=sso_registration_args['sso_region'],
        )
        sso = self._session.create_client('sso', config=client_config)

        sso_account_id, sso_role_name = self._prompt_for_sso_account_and_role(
            sso, sso_token
        )
        configured_for_aws_credentials = all((sso_account_id, sso_role_name))

        self._prompt_for_cli_default_region()
        self._prompt_for_cli_output_format()

        profile_name = self._prompt_for_profile(sso_account_id, sso_role_name)

        self._write_new_config(profile_name)
        self._print_conclusion(configured_for_aws_credentials, profile_name)
        return 0

    def _prompt_for_sso_registration_args(self):
        sso_session = self._sso_session_prompter.prompt_for_sso_session(
            required=False
        )
        if sso_session is None:
            self._warn_configuring_using_legacy_format()
            return self._prompt_for_registration_args_with_legacy_format()
        else:
            self._set_sso_session_in_profile_config(sso_session)
            if sso_session in self._sso_sessions:
                return self._get_sso_registration_args_from_sso_config(
                    sso_session
                )
            else:
                return self._prompt_for_registration_args_for_new_sso_session(
                    sso_session=sso_session
                )

    def _prompt_for_registration_args_with_legacy_format(self):
        self._store_sso_session_prompter_answers_to_profile_config()
        self._set_sso_session_defaults_from_profile_config()
        start_url, sso_region = self._prompt_for_sso_start_url_and_sso_region()
        return {'start_url': start_url, 'sso_region': sso_region}

    def _get_sso_registration_args_from_sso_config(self, sso_session):
        sso_config = self._get_sso_session_config(sso_session)
        return {
            'session_name': sso_session,
            'start_url': sso_config['sso_start_url'],
            'sso_region': sso_config['sso_region'],
            'registration_scopes': sso_config.get('registration_scopes'),
        }

    def _prompt_for_registration_args_for_new_sso_session(self, sso_session):
        self._set_sso_session_defaults_from_profile_config()
        start_url, sso_region = self._prompt_for_sso_start_url_and_sso_region()
        scopes = (
            self._sso_session_prompter.prompt_for_sso_registration_scopes()
        )
        return {
            'session_name': sso_session,
            'start_url': start_url,
            'sso_region': sso_region,
            'registration_scopes': scopes,
            'force_refresh': True,
        }

    def _store_sso_session_prompter_answers_to_profile_config(self):
        self._sso_session_prompter.sso_session_config = (
            self._new_profile_config_values
        )

    def _set_sso_session_in_profile_config(self, sso_session):
        self._new_profile_config_values['sso_session'] = sso_session

    def _set_sso_session_defaults_from_profile_config(self):
        if 'sso_start_url' in self._profile_config:
            self._sso_session_prompter.sso_session_config['sso_start_url'] = (
                self._profile_config['sso_start_url']
            )
        if 'sso_region' in self._profile_config:
            self._sso_session_prompter.sso_session_config['sso_region'] = (
                self._profile_config['sso_region']
            )

    def _prompt_for_sso_start_url_and_sso_region(self):
        start_url = self._sso_session_prompter.prompt_for_sso_start_url()
        sso_region = self._sso_session_prompter.prompt_for_sso_region()
        return start_url, sso_region

    def _warn_configuring_using_legacy_format(self):
        uni_print(
            f'{colorama.Style.BRIGHT}WARNING: Configuring using legacy format '
            f'(e.g. without an SSO session).\n'
            f'Consider re-running "configure sso" command and providing '
            f'a session name.\n{colorama.Style.RESET_ALL}'
        )

    def _prompt_for_sso_account_and_role(self, sso, sso_token):
        sso_account_id = None
        sso_role_name = None
        try:
            sso_account_id = self._prompt_for_account(sso, sso_token)
            sso_role_name = self._prompt_for_role(
                sso, sso_token, sso_account_id
            )
        except sso.exceptions.UnauthorizedException:
            uni_print(
                'Unable to list AWS accounts and/or roles. '
                'Skipping configuring AWS credential provider for profile.\n'
            )
        return sso_account_id, sso_role_name

    def _write_new_config(self, profile):
        if self._new_profile_config_values:
            profile_section = profile_to_section(profile)
            self._update_section(
                profile_section, self._new_profile_config_values
            )
        if self._sso_session_prompter.sso_session:
            self._write_sso_configuration()

    def _print_conclusion(self, configured_for_aws_credentials, profile_name):
        if configured_for_aws_credentials:
            if profile_name.lower() == 'default':
                msg = (
                    'The AWS CLI is now configured to use the default profile.\n'
                    'Run the following command to verify your configuration:\n\n'
                    'aws sts get-caller-identity\n'
                )
            else:
                msg = (
                    'To use this profile, specify the profile name using '
                    '--profile, as shown:\n\n'
                    'aws sts get-caller-identity --profile {}\n'
                )
        else:
            msg = 'Successfully configured SSO for profile: {}\n'
        uni_print(msg.format(profile_name))


class ConfigureSSOSessionCommand(BaseSSOConfigurationCommand):
    NAME = 'sso-session'
    SYNOPSIS = 'aws configure sso-session'
    DESCRIPTION = (
        'The ``aws configure sso-session`` command interactively prompts for '
        'the configuration values required to create a SSO session. '
        'The SSO session can then be associated to a profile to retrieve '
        'SSO access tokens and AWS credentials.\n\n'
        f'{_CMD_PROMPT_USAGE}'
        f'{_CONFIG_EXTRA_INFO}'
    )

    def _run_main(self, parsed_args, parsed_globals):
        super()._run_main(parsed_args, parsed_globals)
        self._sso_session_prompter.prompt_for_sso_session()
        self._sso_session_prompter.prompt_for_sso_start_url()
        self._sso_session_prompter.prompt_for_sso_region()
        self._sso_session_prompter.prompt_for_sso_registration_scopes()
        self._write_sso_configuration()
        self._print_configuration_success()
        return 0

    def _print_configuration_success(self):
        sso_session = self._sso_session_prompter.sso_session
        uni_print(
            f'\nCompleted configuring SSO session: {sso_session}\n'
            f'Run the following to login and refresh access token for '
            f'this session:\n\n'
            f'aws sso login --sso-session {sso_session}\n'
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import base64
import logging
from collections import namedtuple

from awscrt import checksums as crt_checksums
from botocore.httpchecksum import _CHECKSUM_CLS
from s3transfer.exceptions import S3DownloadChecksumError

logger = logging.getLogger(__name__)


CrcCombineInfo = namedtuple('CrcCombineInfo', ['combine_fn', 'byte_length'])


PartChecksum = namedtuple('PartChecksum', ['crc_int', 'data_length'])


FullObjectChecksum = namedtuple(
    'FullObjectChecksum', ['algorithm', 'expected_b64']
)


_CRC_COMBINE_FUNCTIONS = {
    'crc32': CrcCombineInfo(crt_checksums.combine_crc32, 4),
    'crc32c': CrcCombineInfo(crt_checksums.combine_crc32c, 4),
    'crc64nvme': CrcCombineInfo(crt_checksums.combine_crc64nvme, 8),
}


_CHECKSUM_KEY_TO_ALGORITHM = {
    'ChecksumCRC32': 'crc32',
    'ChecksumCRC32C': 'crc32c',
    'ChecksumCRC64NVME': 'crc64nvme',
}


def resolve_full_object_checksum(response):
    if response.get('ChecksumType', '').upper() != 'FULL_OBJECT':
        return None
    for key, algorithm in _CHECKSUM_KEY_TO_ALGORITHM.items():
        value = response.get(key)
        if value:
            return FullObjectChecksum(algorithm=algorithm, expected_b64=value)
    return None


def create_checksum_for_algorithm(algorithm):
    if checksum_cls := _CHECKSUM_CLS.get(algorithm):
        return checksum_cls()
    return None


class FullObjectChecksumCombiner:
    def __init__(self, algorithm, num_parts, expected_b64=None):
        self._algorithm = algorithm
        self._expected_b64 = expected_b64
        self._num_parts = num_parts
        self._combine_info = _CRC_COMBINE_FUNCTIONS[algorithm]
        self._parts = {}
        self._combined_bytes = None

    @property
    def algorithm(self):
        return self._algorithm

    def register_part(self, part_index, checksum, data_length):
        crc_int = int.from_bytes(checksum.digest(), byteorder='big')
        self._parts[part_index] = PartChecksum(crc_int, data_length)

    def combine_and_validate(self):
        if len(self._parts) != self._num_parts:
            logger.debug(
                f'Skipping full object checksum validation. '
                f'Expected {self._num_parts} parts but only '
                f'{len(self._parts)} were registered.'
            )
            return
        combined_bytes = self._get_combined_bytes()
        combined_b64 = base64.b64encode(combined_bytes).decode('ascii')
        expected_bytes = base64.b64decode(self._expected_b64)
        if combined_bytes != expected_bytes:
            raise S3DownloadChecksumError(
                f'Expected full object checksum '
                f'({self._algorithm}) {self._expected_b64} did not match '
                f'combined checksum: {combined_b64}'
            )
        logger.debug(
            f'Full object {self._algorithm} checksum validated: {combined_b64}'
        )

    @property
    def combined_b64(self):
        combined_bytes = self._get_combined_bytes()
        return base64.b64encode(combined_bytes).decode('ascii')

    def _get_combined_bytes(self):
        if self._combined_bytes is not None:
            return self._combined_bytes
        crc = self._parts[0].crc_int
        for i in range(1, self._num_parts):
            part = self._parts[i]
            crc = self._combine_info.combine_fn(
                crc, part.crc_int, part.data_length
            )
        self._combined_bytes = crc.to_bytes(
            self._combine_info.byte_length, byteorder='big'
        )
        return self._combined_bytes



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import shutil
import tempfile

import ruamel.yaml as yaml

from awscli.testutils import BaseAWSCommandParamsTest, capture_input


class TestPut(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestPut, self).setUp()
        self.parsed_response = {}
        self.tempdir = tempfile.mkdtemp()
        self.original_tag_handlers = yaml.YAML(
            typ='safe'
        ).constructor.yaml_constructors.copy()

    def tearDown(self):
        super(TestPut, self).tearDown()
        shutil.rmtree(self.tempdir)
        # This line looks wrong, right?  Well... the "yaml_constructors"
        # is actually a class attribute that's shared across *all* of the
        # yaml.YAML() instances.  Because the ddb customization registers
        # a custom handler for the binary and float types, this will break
        # other code (and tests) that rely on the default behavior.  So
        # to fix this, we put back the original tag handlers that we saved
        # in self.original_tag_handlers to ensure we handle binary and
        # float types correctly.
        yaml.YAML(typ='safe').constructor.yaml_constructors.update(
            self.original_tag_handlers
        )

    def assert_yaml_response_equal(self, response, expected):
        with self.assertRaises(ValueError):
            json.loads(response)
        loaded = yaml.safe_load(response)
        self.assertEqual(loaded, expected)

    def test_simple_put(self):
        command = ['ddb', 'put', 'mytable', '{foo: bar}']
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"S": "bar"}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)
        operations_called = [o[0].name for o in self.operations_called]
        self.assertEqual(operations_called, ['PutItem'])

    def test_batch_write(self):
        command = ['ddb', 'put', 'mytable', '[{foo: bar}, {foo: bar}]']
        expected_params = {
            'ReturnConsumedCapacity': 'NONE',
            'RequestItems': {
                'mytable': [
                    {'PutRequest': {'Item': {'foo': {'S': 'bar'}}}},
                    {'PutRequest': {'Item': {'foo': {'S': 'bar'}}}},
                ]
            },
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)
        operations_called = [o[0].name for o in self.operations_called]
        self.assertEqual(operations_called, ['BatchWriteItem'])

    def test_batch_write_multiple_batches(self):
        items = ', '.join(["{foo: bar}" for _ in range(40)])
        command = ['ddb', 'put', 'mytable', '[%s]' % items]
        self.run_cmd(command, expected_rc=0)
        operations_called = [o[0].name for o in self.operations_called]
        self.assertEqual(
            operations_called, ['BatchWriteItem', 'BatchWriteItem']
        )

        first_params = self.operations_called[0][1]
        num_items = len(first_params['RequestItems']['mytable'])
        self.assertEqual(num_items, 25)
        self.assertEqual(first_params.get('ReturnConsumedCapacity'), 'NONE')

        second_params = self.operations_called[1][1]
        num_items = len(second_params['RequestItems']['mytable'])
        self.assertEqual(num_items, 15)
        self.assertEqual(second_params.get('ReturnConsumedCapacity'), 'NONE')

    def test_batch_write_unprocessed_items(self):
        self.parsed_responses = [
            {
                'UnprocessedItems': {
                    'mytable': [
                        {'PutRequest': {'Item': {'foo': {'S': 'bar'}}}},
                    ]
                }
            },
            {
                'UnprocessedItems': {
                    'mytable': [
                        {'PutRequest': {'Item': {'foo': {'S': 'bar'}}}},
                    ]
                }
            },
            {},
        ]

        items = ', '.join(["{foo: bar}" for _ in range(40)])
        command = ['ddb', 'put', 'mytable', '[%s]' % items]
        self.run_cmd(command, expected_rc=0)
        operations_called = [o[0].name for o in self.operations_called]
        self.assertEqual(
            operations_called,
            [
                'BatchWriteItem',
                'BatchWriteItem',
                'BatchWriteItem',
            ],
        )

        first_params = self.operations_called[0][1]
        num_items = len(first_params['RequestItems']['mytable'])
        self.assertEqual(num_items, 25)

        second_params = self.operations_called[1][1]
        num_items = len(second_params['RequestItems']['mytable'])
        self.assertEqual(num_items, 16)

        third_params = self.operations_called[2][1]
        num_items = len(third_params['RequestItems']['mytable'])
        self.assertEqual(num_items, 1)

    def test_put_with_condition(self):
        command = [
            'ddb',
            'put',
            'mytable',
            '{foo: bar}',
            '--condition',
            'attribute_exists(foo)',
        ]
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'ConditionExpression': 'attribute_exists(#n0)',
            'ExpressionAttributeNames': {'#n0': 'foo'},
            'Item': {"foo": {"S": "bar"}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)

    def test_batch_write_with_condition(self):
        command = [
            'ddb',
            'put',
            'mytable',
            '[{foo: bar}, {foo: bar}]',
            '--condition',
            'attribute_exists(foo)',
        ]
        _, stderr, _ = self.assert_params_for_cmd(command, expected_rc=252)
        self.assertIn('--condition is not supported', stderr)

    def test_load_items_from_file(self):
        filename = os.path.join(self.tempdir, 'items.yml')
        with open(filename, 'w') as f:
            f.write('{foo: bar}\n')

        command = ['ddb', 'put', 'mytable', 'file://%s' % filename]
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"S": "bar"}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)

    def test_load_items_from_stdin(self):
        command = ['ddb', 'put', 'mytable', '-']
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"S": "bar"}},
        }
        with capture_input(b'{foo: bar}'):
            self.assert_params_for_cmd(command, expected_params, expected_rc=0)

    def test_put_bytes(self):
        command = ['ddb', 'put', 'mytable', '{foo: !!binary "4pyT"}']
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"B": b'\xe2\x9c\x93'}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)

    def test_put_int(self):
        command = ['ddb', 'put', 'mytable', '{foo: 1}']
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"N": "1"}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)

    def test_put_float(self):
        command = ['ddb', 'put', 'mytable', '{foo: 1.1}']
        expected_params = {
            'TableName': 'mytable',
            'ReturnConsumedCapacity': 'NONE',
            'Item': {"foo": {"N": "1.1"}},
        }
        self.assert_params_for_cmd(command, expected_params, expected_rc=0)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestRemoveTags(BaseAWSCommandParamsTest):
    prefix = 'elb remove-tags'

    def test_shorthand(self):
        command = self.prefix + ' --load-balancer-names foo'
        command += ' --tags bar baz'
        expected_params = {
            'LoadBalancerNames': ['foo'],
            'Tags': [{'Key': 'bar'}, {'Key': 'baz'}],
        }
        self.assert_params_for_cmd(command, expected_params)



#!/usr/bin/env python
# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy
import re

from awscli.testutils import BaseAWSCommandParamsTest

# file is gone in python3, so instead IOBase must be used.
# Given this test module is the only place that cares about
# this type check, we do the check directly in this test module.
try:
    file_type = file
except NameError:
    import io

    file_type = io.IOBase


TAGSET = """{"TagSet":[{"Key":"key1","Value":"value1"},{"Key":"key2","Value":"value2"}]}"""


class TestPutBucketTagging(BaseAWSCommandParamsTest):
    prefix = 's3api put-bucket-tagging'

    def setUp(self):
        super(TestPutBucketTagging, self).setUp()
        self.payload = None

    def test_simple(self):
        cmdline = self.prefix
        cmdline += ' --bucket mybucket'
        cmdline += ' --tagging %s' % TAGSET
        expected = {
            'Bucket': 'mybucket',
            'Tagging': {
                'TagSet': [
                    {'Key': 'key1', 'Value': 'value1'},
                    {'Key': 'key2', 'Value': 'value2'},
                ]
            },
        }
        self.assert_params_for_cmd(cmdline, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import contextlib
import json

from botocore.config import Config
from botocore.exceptions import ClientError

from tests import BaseSessionTest, ClientHTTPStubber, mock
from tests.functional.botocore.test_useragent import (
    get_captured_ua_strings,
    parse_registered_feature_ids,
)


class TestRetry(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.sleep_patch = mock.patch('time.sleep')
        self.sleep_patch.start()

    def tearDown(self):
        super().tearDown()
        self.sleep_patch.stop()

    @contextlib.contextmanager
    def assert_will_retry_n_times(
        self, client, num_retries, status=500, body=b'{}'
    ):
        num_responses = num_retries + 1
        if not isinstance(body, bytes):
            body = json.dumps(body).encode('utf-8')
        with ClientHTTPStubber(client) as http_stubber:
            for _ in range(num_responses):
                http_stubber.add_response(status=status, body=body)
            with self.assertRaisesRegex(
                ClientError, f'reached max retries: {num_retries}'
            ):
                yield
            self.assertEqual(len(http_stubber.requests), num_responses)

    def _get_feature_id_lists_from_retries(self, client):
        with ClientHTTPStubber(client) as http_stubber:
            # Add two failed responses followed by a success
            http_stubber.add_response(status=502, body=b'{}')
            http_stubber.add_response(status=502, body=b'{}')
            http_stubber.add_response(status=200, body=b'{}')
            client.list_tables()
        ua_strings = get_captured_ua_strings(http_stubber)
        return [
            parse_registered_feature_ids(ua_string) for ua_string in ua_strings
        ]

    def create_client_with_retry_config(
        self, service, retry_mode='standard', max_attempts=None
    ):
        retries = {'mode': retry_mode}
        if max_attempts is not None:
            retries['max_attempts'] = max_attempts
        client = self.session.create_client(
            service, self.region, config=Config(retries=retries)
        )
        return client

    def test_can_override_only_max_attempts(self):
        client = self.session.create_client(
            'dynamodb', self.region, config=Config(retries={'max_attempts': 2})
        )
        with self.assert_will_retry_n_times(client, 1):
            client.list_tables()

    def test_do_not_attempt_retries(self):
        client = self.create_client_with_retry_config(
            'dynamodb', max_attempts=1
        )
        with self.assert_will_retry_n_times(client, 0):
            client.list_tables()

    def test_setting_max_attempts_does_not_set_for_other_clients(self):
        # Make one client with max attempts configured.
        self.create_client_with_retry_config('dynamodb', max_attempts=1)

        # Make another client that has no custom retry configured.
        client = self.session.create_client('codecommit', self.region)
        # It should use the default max retries, which should be two retries
        # for this service.
        with self.assert_will_retry_n_times(client, 2):
            client.list_repositories()

    def test_set_max_attempts_on_session(self):
        self.session.set_default_client_config(
            Config(retries={'max_attempts': 2})
        )
        # Max attempts should be inherited from the session.
        client = self.session.create_client('codecommit', self.region)
        with self.assert_will_retry_n_times(client, 1):
            client.list_repositories()

    def test_can_clobber_max_attempts_on_session(self):
        self.session.set_default_client_config(
            Config(retries={'max_attempts': 5})
        )
        # Max attempts should override the session's configured max attempts.
        client = self.session.create_client(
            'codecommit',
            self.region,
            config=Config(retries={'max_attempts': 3}),
        )
        with self.assert_will_retry_n_times(client, 2):
            client.list_repositories()

    def test_standard_mode_has_default_3_retries(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        with self.assert_will_retry_n_times(client, 2):
            client.list_tables()

    def test_standard_mode_can_configure_max_attempts(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard', max_attempts=5
        )
        with self.assert_will_retry_n_times(client, 4):
            client.list_tables()

    def test_no_retry_needed_standard_mode(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'{}')
            client.list_tables()

    def test_standard_mode_retry_throttling_error(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        error_body = {"__type": "ThrottlingException", "message": "Error"}
        with self.assert_will_retry_n_times(
            client, 2, status=400, body=error_body
        ):
            client.list_tables()

    def test_standard_mode_retry_transient_error(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        with self.assert_will_retry_n_times(client, 2, status=502):
            client.list_tables()

    def test_can_exhaust_default_retry_quota(self):
        # Quota of 500 / 5 retry costs == 100 retry attempts
        # 100 retry attempts / 2 retries per API call == 50 client calls
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        for i in range(50):
            with self.assert_will_retry_n_times(client, 2, status=502):
                client.list_tables()
        # Now on the 51th attempt we should see quota errors, which we can
        # verify by looking at the request metadata.
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=502, body=b'{}')
            with self.assertRaises(ClientError) as e:
                client.list_tables()
        self.assertTrue(
            e.exception.response['ResponseMetadata'].get('RetryQuotaReached')
        )

    def test_user_agent_has_standard_mode_feature_id(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='standard'
        )
        feature_lists = self._get_feature_id_lists_from_retries(client)
        # Confirm all requests register `'RETRY_MODE_STANDARD': 'E'`
        assert all('E' in feature_list for feature_list in feature_lists)

    def test_user_agent_has_adaptive_mode_feature_id(self):
        client = self.create_client_with_retry_config(
            'dynamodb', retry_mode='adaptive'
        )
        feature_lists = self._get_feature_id_lists_from_retries(client)
        # Confirm all requests register `'RETRY_MODE_ADAPTIVE': 'F'`
        assert all('F' in feature_list for feature_list in feature_lists)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from awscli.alias import AliasLoader, register_alias_commands
from awscli.customizations.commands import BasicCommand
from awscli.testutils import (
    BaseAWSCommandParamsTest,
    FileCreator,
    skip_if_windows,
)


# We need to test that aliases work with custom BasicCommand,
# but the actual custom commands in the CLI can be complex and
# don't want the alias tests to fail if they are updated with new API
# calls they invoke.  This custom command is a simple wrapper to the
# aws resourcegroupstaggingapi get-resources command.
class CustomTestCommand(BasicCommand):
    NAME = 'resources'
    ARG_TABLE = [
        {'name': 'resource-types', 'action': 'store', 'required': False},
        {'name': 'tags', 'action': 'store', 'required': False},
    ]

    def _run_main(self, parsed_args, parsed_globals):
        client = self._session.create_client('resourcegroupstaggingapi')
        kwargs = {}
        if parsed_args.resource_types is not None:
            kwargs['ResourceTypeFilters'] = [parsed_args.resource_types]
        if parsed_args.tags is not None:
            key, value = parsed_args.tags.split('=')
            kwargs['TagFilters'] = [{'Key': key, 'Values': [value]}]
        client.get_resources(**kwargs)


class TestAliases(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestAliases, self).setUp()
        self.files = FileCreator()
        self.alias_file = self.files.create_file('alias', '[toplevel]\n')
        self.driver.alias_loader = AliasLoader(self.alias_file)
        self.event_emitter = self.driver.session.get_component('event_emitter')
        # Alias injection is part of the built-in handler chain which defaults
        # to the real ~/.aws/cli/alias file.  We need to unregister the default
        # injector so we can swap in our own version that points to a
        # test file.
        self.event_emitter.unregister(
            'building-command-table', unique_id='cli-alias-injector'
        )
        register_alias_commands(
            self.event_emitter, alias_filename=self.alias_file
        )

    def tearDown(self):
        super(TestAliases, self).tearDown()
        self.files.remove_all()

    def add_alias(self, alias_name, alias_value):
        with open(self.alias_file, 'a+') as f:
            f.write('%s = %s\n' % (alias_name, alias_value))

    def assert_single_operation_called(
        self, cmdline, service_name, operation_name, params
    ):
        self.assert_params_for_cmd(cmdline, params)
        operations_called = self.operations_called
        self.assertEqual(len(operations_called), 1)
        single_op = operations_called[0]
        self.assertEqual(
            single_op[0].service_model.service_name,
            service_name,
        )
        self.assertEqual(single_op[0].name, operation_name)

    def test_subcommand_alias(self):
        self.add_alias('my-alias', 'ec2 describe-regions')
        cmdline = 'my-alias'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={},
        )

    def test_subcommand_alias_with_additonal_params(self):
        self.add_alias(
            'my-alias', 'ec2 describe-regions --region-names us-east-1'
        )
        cmdline = 'my-alias'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-east-1']},
        )

    def test_subcommand_alias_then_additonal_params(self):
        self.add_alias('my-alias', 'ec2')
        cmdline = 'my-alias describe-regions --region-names us-east-1'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-east-1']},
        )

    def test_subcommand_alias_with_global_params(self):
        self.add_alias(
            'my-alias',
            'ec2 describe-regions --query Regions[].RegionName --output text',
        )
        self.parsed_responses = [
            {
                'Regions': [
                    {
                        'Endpoint': 'ec2.us-east-1.amazonaws.com',
                        'RegionName': 'us-east-1',
                    }
                ]
            }
        ]
        cmdline = 'my-alias'
        stdout, _, _ = self.assert_params_for_cmd(cmdline, {})
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].service_model.service_name, 'ec2'
        )
        self.assertEqual(self.operations_called[0][0].name, 'DescribeRegions')
        self.assertEqual(stdout.strip(), 'us-east-1')

    def test_subcommand_alias_then_global_params(self):
        self.add_alias('my-alias', 'ec2 describe-regions')
        self.parsed_responses = [
            {
                'Regions': [
                    {
                        'Endpoint': 'ec2.us-east-1.amazonaws.com',
                        'RegionName': 'us-east-1',
                    }
                ]
            }
        ]
        cmdline = 'my-alias '
        cmdline += '--query=Regions[].RegionName '
        cmdline += '--output=text'
        stdout, _, _ = self.assert_params_for_cmd(cmdline, {})
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].service_model.service_name, 'ec2'
        )
        self.assertEqual(self.operations_called[0][0].name, 'DescribeRegions')
        self.assertEqual(stdout.strip(), 'us-east-1')

    def test_global_params_then_subcommand_alias(self):
        self.add_alias('my-alias', 'ec2 describe-regions')
        self.parsed_responses = [
            {
                'Regions': [
                    {
                        'Endpoint': 'ec2.us-east-1.amazonaws.com',
                        'RegionName': 'us-east-1',
                    }
                ]
            }
        ]
        cmdline = '--query=Regions[].RegionName '
        cmdline += '--output=text '
        cmdline += 'my-alias'
        stdout, _, _ = self.assert_params_for_cmd(cmdline, {})
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].service_model.service_name, 'ec2'
        )
        self.assertEqual(self.operations_called[0][0].name, 'DescribeRegions')
        self.assertEqual(stdout.strip(), 'us-east-1')

    def test_alias_overrides_builtin_command(self):
        self.add_alias('ec2', 's3api')
        cmdline = 'ec2 list-buckets'
        self.assert_params_for_cmd(cmdline, {})
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(self.operations_called[0][0].name, 'ListBuckets')

    def test_alias_proxies_to_shadowed_command(self):
        self.add_alias('ec2', 'ec2')
        cmdline = 'ec2 describe-regions'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={},
        )

    def test_alias_chaining(self):
        self.add_alias('base-alias', 'ec2 describe-regions')
        self.add_alias('wrapper-alias', 'base-alias --region-names us-east-1')
        cmdline = 'wrapper-alias'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-east-1']},
        )

    def test_alias_chaining_with_globals(self):
        self.add_alias('base-alias', 'ec2 describe-regions')
        self.add_alias(
            'wrapper-alias',
            'base-alias --query Regions[].RegionName --output text',
        )
        cmdline = 'wrapper-alias'
        self.parsed_responses = [
            {
                'Regions': [
                    {
                        'Endpoint': 'ec2.us-east-1.amazonaws.com',
                        'RegionName': 'us-east-1',
                    }
                ]
            }
        ]
        stdout, _, _ = self.assert_params_for_cmd(cmdline, {})
        self.assertEqual(len(self.operations_called), 1)
        self.assertEqual(
            self.operations_called[0][0].service_model.service_name, 'ec2'
        )
        self.assertEqual(self.operations_called[0][0].name, 'DescribeRegions')
        self.assertEqual(stdout.strip(), 'us-east-1')

    def test_external_alias(self):
        # The external alias is tested by using mkdir; a command that
        # is universal for the various OS's we support
        directory_to_make = os.path.join(self.files.rootdir, 'newdir')
        self.add_alias('mkdir', '!mkdir %s' % directory_to_make)
        self.run_cmd('mkdir')
        self.assertTrue(os.path.isdir(directory_to_make))

    def test_external_alias_then_additonal_args(self):
        # The external alias is tested by using mkdir; a command that
        # is universal for the various OS's we support
        directory_to_make = os.path.join(self.files.rootdir, 'newdir')
        self.add_alias('mkdir', '!mkdir')
        self.run_cmd('mkdir %s' % directory_to_make)
        self.assertTrue(os.path.isdir(directory_to_make))

    def test_external_alias_with_quoted_arguments(self):
        directory_to_make = os.path.join(self.files.rootdir, 'new dir')
        self.add_alias('mkdir', '!mkdir')
        self.run_cmd(['mkdir', directory_to_make])
        self.assertTrue(os.path.isdir(directory_to_make))

    @skip_if_windows('Windows does not support BASH functions')
    def test_external_alias_with_wrapper_bash_function(self):
        # The external alias is tested by using mkdir; a command that
        # is universal for the various OS's we support
        directory_to_make = os.path.join(self.files.rootdir, 'newdir')
        self.add_alias('mkdir', '!f() { mkdir "${1}"; }; f')
        self.run_cmd('mkdir %s' % directory_to_make)
        self.assertTrue(os.path.isdir(directory_to_make))

    def test_operation_level_alias_with_additonal_params(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2]\n')
            f.write('regions = describe-regions --region-names us-west-2\n')

        cmdline = 'ec2 regions'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-west-2']},
        )

    def test_can_shadow_subcommand_alias(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2]\n')
            f.write(
                'describe-regions = describe-regions --region-names us-west-2\n'
            )

        cmdline = 'ec2 describe-regions'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-west-2']},
        )

    def test_multi_nested_command_alias(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2 wait]\n')
            f.write('vpc-ready = vpc-available\n')

        cmdline = 'ec2 wait vpc-ready'
        self.parsed_response = {
            'Vpcs': [{'State': 'available'}],
        }
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeVpcs',
            params={},
        )

    def test_can_extend_subcommand_internal_aliases(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2]\n')
            f.write(
                'regions = describe-regions ' '--query Regions[].RegionName\n'
            )
        cmdline = 'ec2 regions --region-names us-west-2'
        self.assert_single_operation_called(
            cmdline,
            service_name='ec2',
            operation_name='DescribeRegions',
            params={'RegionNames': ['us-west-2']},
        )

    def test_operation_level_external_alias(self):
        directory_to_make = os.path.join(self.files.rootdir, 'newdir')
        with open(self.alias_file, 'a+') as f:
            f.write('[command ec2]\n')
            f.write('mkdir = !mkdir\n')
        self.run_cmd('ec2 mkdir %s' % directory_to_make)
        self.assertTrue(os.path.isdir(directory_to_make))

    def test_can_create_bag_of_options_alias(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command cloudformation list-stacks]\n')
            f.write(
                'created = --stack-status-filter CREATE_COMPLETE '
                '--query StackSummaries[].[StackName,StackStatus] '
                '--output text\n'
            )

        cmdline = 'cloudformation list-stacks created'
        self.assert_single_operation_called(
            cmdline,
            service_name='cloudformation',
            operation_name='ListStacks',
            params={'StackStatusFilter': ['CREATE_COMPLETE']},
        )

    def test_can_support_multiple_alias_expansions(self):
        with open(self.alias_file, 'a+') as f:
            f.write(
                'cfn = cloudformation\n'
                '[command cloudformation]\n'
                'stacks = list-stacks\n'
                '[command cloudformation list-stacks]\n'
                'created = --stack-status-filter CREATE_COMPLETE '
                '--query StackSummaries[].[StackName,StackStatus] '
                '--output text\n'
            )

        cmdline = 'cfn stacks created'
        self.assert_single_operation_called(
            cmdline,
            service_name='cloudformation',
            operation_name='ListStacks',
            params={'StackStatusFilter': ['CREATE_COMPLETE']},
        )

    def test_can_merge_explicit_and_alias_local_params(self):
        section = '[command resourcegroupstaggingapi get-resources]\n'
        alias = (
            'mysvc = --tag-filters Key=foo,Values=bar '
            'Key=bar,Values=baz --query ResourceTagMappingList[].[ResourceARN]'
            ' --output text\n'
        )
        with open(self.alias_file, 'a+') as f:
            f.write(section)
            f.write(alias)
        # If we add additional operatio-specific params such as
        # "--resource-type-filters ecs", then we should merge
        # those values with the params specified in the alias definition.
        cmdline = (
            'resourcegroupstaggingapi get-resources mysvc '
            '--resource-type-filters ecs'
        )
        self.assert_single_operation_called(
            cmdline,
            service_name='resourcegroupstaggingapi',
            operation_name='GetResources',
            params={
                'TagFilters': [
                    {'Key': 'foo', 'Values': ['bar']},
                    {'Key': 'bar', 'Values': ['baz']},
                ],
                'ResourceTypeFilters': ['ecs'],
            },
        )

    def test_can_handle_bag_of_options_with_required_args(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command iam create-user]\n')
            f.write('test-user = --user-name test-user\n')
        # The current behavior without aliases is that `--user-name`
        # is a required parameter.  So without changing anything
        # if you ran "aws iam create-user test-user", the parser
        # would complain that the `--user-name` arg is required.
        # However, the `--user-name` is specified, it's just in the
        # `test-user` alias.  So we have to ensure that we're able
        # to delegate to the alias in the hopes that it can provide
        # the required value.  FWIW we had to do something similar
        # in the `--generate-cli-skeleton` flow.
        self.assert_single_operation_called(
            'iam create-user test-user',
            service_name='iam',
            operation_name='CreateUser',
            params={'UserName': 'test-user'},
        )

    def test_can_handle_bag_of_options_with_custom_command(self):
        self.event_emitter.register(
            'building-command-table.resourcegroupstaggingapi',
            CustomTestCommand.add_command,
        )
        with open(self.alias_file, 'a+') as f:
            f.write('[command resourcegroupstaggingapi resources]\n')
            f.write('mycluster = --resource-types ecs:cluster\n')
        cmdline = 'resourcegroupstaggingapi resources mycluster'
        self.assert_single_operation_called(
            cmdline,
            service_name='resourcegroupstaggingapi',
            operation_name='GetResources',
            params={'ResourceTypeFilters': ['ecs:cluster']},
        )

    def test_can_handle_custom_bag_of_options_and_overrides(self):
        self.event_emitter.register(
            'building-command-table.resourcegroupstaggingapi',
            CustomTestCommand.add_command,
        )
        with open(self.alias_file, 'a+') as f:
            f.write('[command resourcegroupstaggingapi resources]\n')
            f.write('mycluster = --resource-types ecs:cluster\n')
        cmdline = 'resourcegroupstaggingapi resources mycluster --tags foo=bar'
        # We should merge the cmdline params from the 'mycluster' alias along
        # with the explicitly provided --tags option from cmdline args.
        self.assert_single_operation_called(
            cmdline,
            service_name='resourcegroupstaggingapi',
            operation_name='GetResources',
            params={
                'ResourceTypeFilters': ['ecs:cluster'],
                'TagFilters': [{'Key': 'foo', 'Values': ['bar']}],
            },
        )

    def test_can_handle_multiple_bag_of_options(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ecs describe-tasks]\n')
            f.write('mycluster = --cluster mycluster\n')
            f.write('mytasks = --tasks foo\n')

        cmdline = 'ecs describe-tasks mycluster mytasks'
        self.assert_single_operation_called(
            cmdline,
            service_name='ecs',
            operation_name='DescribeTasks',
            params={'cluster': 'mycluster', 'tasks': ['foo']},
        )

    def test_can_handlle_multiple_bags_of_options_with_overrides(self):
        with open(self.alias_file, 'a+') as f:
            f.write('[command ecs describe-tasks]\n')
            f.write('mycluster = --cluster mycluster\n')
            f.write('mytasks = --tasks foo\n')

        cmdline = 'ecs describe-tasks mycluster mytasks --include TAGS'
        self.assert_single_operation_called(
            cmdline,
            service_name='ecs',
            operation_name='DescribeTasks',
            params={
                'cluster': 'mycluster',
                'tasks': ['foo'],
                'include': ['TAGS'],
            },
        )

    def test_can_invoke_custom_test_command(self):
        # Sanity check to ensure the custom test command we're using
        # works properly so we don't waste time debugging a test because
        # the test command implementation is wrong.
        self.event_emitter.register(
            'building-command-table.resourcegroupstaggingapi',
            CustomTestCommand.add_command,
        )
        cmdline = 'resourcegroupstaggingapi resources --resource-types ecs --tags foo=bar'
        self.assert_single_operation_called(
            cmdline,
            service_name='resourcegroupstaggingapi',
            operation_name='GetResources',
            params={
                'ResourceTypeFilters': ['ecs'],
                'TagFilters': [{'Key': 'foo', 'Values': ['bar']}],
            },
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from argparse import Namespace

from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.credentials import Credentials
from botocore.session import Session

import awscli
from awscli.compat import StringIO
from awscli.customizations.codecommit import (
    CodeCommitCommand,
    CodeCommitGetCommand,
)
from awscli.customizations.exceptions import ParamValidationError
from awscli.testutils import StringIOWithFileNo, mock, unittest


class TestCodeCommitCredentialHelper(unittest.TestCase):
    PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=git-codecommit.us-east-1.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    PROTOCOL_HOST_PATH_TRAILING_NEWLINE = (
        'protocol=https\n'
        'host=git-codecommit.us-east-1.amazonaws.com\n'
        'path=/v1/repos/myrepo\n'
    )

    PROTOCOL_HOST_PATH_BLANK_LINE = (
        'protocol=https\n'
        'host=git-codecommit.us-east-1.amazonaws.com\n'
        'path=/v1/repos/myrepo\n\n'
    )

    FIPS_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=git-codecommit-fips.us-east-1.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    VPC_1_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=vpce-0b47ea360adebf88a-jkl88hez.git-codecommit.us-east-1.vpce.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    VPC_2_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=vpce-0b47ea360adebf88a-jkl88hez-us-east-1a.git-codecommit.us-east-1.vpce.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    FIPS_VPC_1_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=vpce-0b47ea360adebf88a-jkl88hez.git-codecommit-fips.us-east-1.vpce.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    FIPS_VPC_2_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=vpce-0b47ea360adebf88a-jkl88hez-us-west-2b.git-codecommit-fips.us-east-1.vpce.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    NO_REGION_PROTOCOL_HOST_PATH = (
        'protocol=https\n'
        'host=git-codecommit.amazonaws.com\n'
        'path=/v1/repos/myrepo'
    )

    NON_AWS_PROTOCOL_HOST_PATH = (
        'protocol=https\n' 'host=mydomain.com\n' 'path=/v1/repos/myrepo'
    )

    MOCK_STDOUT_CLASS = StringIOWithFileNo

    def setUp(self):
        self.credentials = Credentials('access', 'secret')
        self.args = Namespace()
        self.args.ignore_host_check = False
        self.globals = Namespace()
        self.globals.region = 'us-east-1'
        self.globals.verify_ssl = False
        self.session = mock.MagicMock()
        self.session.get_config_variable.return_value = 'us-east-1'
        self.session.get_credentials.return_value = self.credentials

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(PROTOCOL_HOST_PATH))
    def test_generate_credentials(self, stdout_mock):
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(PROTOCOL_HOST_PATH_TRAILING_NEWLINE))
    def test_generate_credentials_trailing_newline(self, stdout_mock):
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(PROTOCOL_HOST_PATH_BLANK_LINE))
    def test_generate_credentials_blank_line(self, stdout_mock):
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(FIPS_PROTOCOL_HOST_PATH))
    def test_generate_credentials_fips_reads_region_from_url(
        self, stdout_mock
    ):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(VPC_1_PROTOCOL_HOST_PATH))
    def test_generate_credentials_vpc_reads_region_from_url(self, stdout_mock):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(VPC_2_PROTOCOL_HOST_PATH))
    def test_generate_credentials_vpc_2_reads_region_from_url(
        self, stdout_mock
    ):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(FIPS_VPC_1_PROTOCOL_HOST_PATH))
    def test_generate_credentials_fips_vpc_1_reads_region_from_url(
        self, stdout_mock
    ):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(FIPS_VPC_2_PROTOCOL_HOST_PATH))
    def test_generate_credentials_fips_vpc_2_reads_region_from_url(
        self, stdout_mock
    ):
        self.globals.region = None
        self.session.get_config_variable.return_value = None
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(NO_REGION_PROTOCOL_HOST_PATH))
    def test_generate_credentials_reads_region_from_session(self, stdout_mock):
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(output, 'username={0}\npassword=.+'.format('access'))

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(NON_AWS_PROTOCOL_HOST_PATH))
    def test_does_nothing_for_non_amazon_domain(self, stdout_mock):
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertEqual('', output)

    def test_raises_validation_error_when_not_provided_any_subcommands(self):
        self.get_command = CodeCommitCommand(self.session)
        with self.assertRaises(ParamValidationError):
            self.get_command._run_main(self.args, self.globals)

    @mock.patch('sys.stdout', new_callable=MOCK_STDOUT_CLASS)
    @mock.patch('sys.stdin', StringIO(PROTOCOL_HOST_PATH))
    def test_generate_session_credentials(self, stdout_mock):
        self.credentials = Credentials('access', 'secret', 'token')
        self.session.get_credentials.return_value = self.credentials
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        output = stdout_mock.getvalue().strip()
        self.assertRegex(
            output, 'username={0}%{1}\npassword=.+'.format('access', 'token')
        )

    @mock.patch('sys.stdout', MOCK_STDOUT_CLASS())
    @mock.patch('sys.stdin', StringIO(PROTOCOL_HOST_PATH))
    @mock.patch('botocore.auth.SigV4Auth.string_to_sign')
    @mock.patch('botocore.auth.SigV4Auth.signature')
    def test_generate_credentials_creates_a_valid_request(
        self, signature, string_to_sign
    ):
        self.credentials = Credentials('access', 'secret')
        self.session.get_credentials.return_value = self.credentials
        self.get_command = CodeCommitGetCommand(self.session)
        self.get_command._run_main(self.args, self.globals)
        aws_request = signature.call_args[0][1]
        self.assertEqual('GIT', aws_request.method)
        self.assertEqual(
            'https://git-codecommit.us-east-1.amazonaws.com//v1/repos/myrepo',
            aws_request.url,
        )
        self.assertEqual(
            (
                'GIT\n//v1/repos/myrepo\n\n'
                'host:git-codecommit.us-east-1.amazonaws.com\n\n'
                'host\n'
            ),
            string_to_sign.call_args[0][1],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.compat import StringIO
from awscli.customizations.configure.get import ConfigureGetCommand
from awscli.testutils import unittest

from . import FakeSession


class TestConfigureGetCommand(unittest.TestCase):
    def create_command(self, session):
        stdout = StringIO()
        stderr = StringIO()
        command = ConfigureGetCommand(session, stdout, stderr)
        return stdout, stderr, command

    def test_configure_get_command(self):
        session = FakeSession({})
        session.config['region'] = 'us-west-2'
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['region'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'us-west-2')

    def test_configure_get_command_no_exist(self):
        no_vars_defined = {}
        session = FakeSession(no_vars_defined)
        stream, error_stream, config_get = self.create_command(session)
        rc = config_get(args=['region'], parsed_globals=None)
        rendered = stream.getvalue()
        # If a config value does not exist, we don't print any output.
        self.assertEqual(rendered, '')
        # And we exit with an rc of 1.
        self.assertEqual(rc, 1)

    def test_dotted_get(self):
        session = FakeSession({})
        session.full_config = {'plugins': {'bar': 'true'}}
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['plugins.bar'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'true')

    def test_dotted_get_with_profile(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'emr-dev': {'emr': {'instance_profile': 'my_ip'}}}
        }
        session.config = {'emr': {'instance_profile': 'my_ip'}}
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['emr-dev.emr.instance_profile'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'my_ip')

    def test_get_from_profile(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'testing': {'aws_access_key_id': 'access_key'}}
        }
        stream, error_stream, config_get = self.create_command(session)
        config_get = ConfigureGetCommand(session, stream)
        config_get(
            args=['profile.testing.aws_access_key_id'], parsed_globals=None
        )
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'access_key')

    def test_get_nested_attribute(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'testing': {'s3': {'signature_version': 's3v4'}}}
        }
        stream, error_stream, config_get = self.create_command(session)
        config_get(
            args=['profile.testing.s3.signature_version'], parsed_globals=None
        )
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 's3v4')

    def test_get_nested_attribute_from_default(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'default': {'s3': {'signature_version': 's3v4'}}}
        }
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['default.s3.signature_version'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 's3v4')

    def test_get_nested_attribute_from_default_does_not_exist(self):
        session = FakeSession({})
        session.full_config = {'profiles': {}}
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['default.s3.signature_version'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), '')

    def test_get_nested_attribute_from_implicit_default(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'default': {'s3': {'signature_version': 's3v4'}}}
        }
        stream, error_stream, config_get = self.create_command(session)
        config_get(args=['s3.signature_version'], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 's3v4')

    def test_get_section_returns_error(self):
        session = FakeSession({})
        session.full_config = {
            'profiles': {'default': {'s3': {'signature_version': 's3v4'}}}
        }
        session.config = {'s3': {'signature_version': 's3v4'}}
        stream, error_stream, config_get = self.create_command(session)
        rc = config_get(args=['s3'], parsed_globals=None)
        self.assertEqual(rc, 1)

        error_message = error_stream.getvalue()
        expected_message = (
            'varname (s3) must reference a value, not a section or '
            'sub-section.'
        )
        self.assertEqual(error_message, expected_message)
        self.assertEqual(stream.getvalue(), '')

    def test_get_non_string_returns_error(self):
        # This should never happen, but we handle this case so we should
        # test it.
        session = FakeSession({})
        session.full_config = {'profiles': {'default': {'foo': object()}}}
        stream, error_stream, config_get = self.create_command(session)
        rc = config_get(args=['foo'], parsed_globals=None)
        self.assertEqual(rc, 1)
        self.assertEqual(stream.getvalue(), '')
        self.assertEqual(error_stream.getvalue(), '')


class TestConfigureGetCommandSubSections(unittest.TestCase):
    def create_command(self, session):
        stdout = StringIO()
        stderr = StringIO()
        command = ConfigureGetCommand(session, stdout, stderr)
        return stdout, stderr, command

    def test_get_simple_property(self):
        session = FakeSession({})
        session.full_config = {
            'sso_sessions': {
                'my-session': {
                    'sso_region': 'us-west-2',
                    'sso_start_url': 'https://example.awsapps.com/start',
                }
            }
        }
        stream, error_stream, config_get = self.create_command(session)
        rc = config_get(
            args=['--sso-session', 'my-session', 'sso_region'],
            parsed_globals=None,
        )
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'us-west-2')
        self.assertEqual(rc, 0)

    def test_get_nested_property(self):
        session = FakeSession({})
        session.full_config = {
            'services': {
                'my-services': {
                    's3': {
                        'endpoint_url': 'http://localhost:4566',
                    }
                }
            }
        }
        stream, error_stream, config_get = self.create_command(session)
        rc = config_get(
            args=['--services', 'my-services', 's3.endpoint_url'],
            parsed_globals=None,
        )
        rendered = stream.getvalue()
        self.assertEqual(rendered.strip(), 'http://localhost:4566')
        self.assertEqual(rc, 0)

    def test_non_existent_property_returns_exit_code_1(self):
        session = FakeSession({})
        session.full_config = {
            'sso_sessions': {'my-session': {'sso_region': 'us-west-2'}}
        }
        stream, _, config_get = self.create_command(session)
        rc = config_get(
            args=['--sso-session', 'my-session', 'non_existent_property'],
            parsed_globals=None,
        )

        self.assertEqual(rc, 1)
        self.assertEqual(stream.getvalue(), '')



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestRebootWorkspaces(BaseAWSCommandParamsTest):
    prefix = 'workspaces reboot-workspaces'

    def test_shorthand(self):
        command = self.prefix + ' --reboot-workspace-requests id1 id2 id3'
        expected_params = {
            'RebootWorkspaceRequests': [
                {"WorkspaceId": "id1"},
                {"WorkspaceId": "id2"},
                {"WorkspaceId": "id3"},
            ]
        }
        self.assert_params_for_cmd(command, expected_params)
